PROGRAM OrnekMPI
include 'mpif.h'

integer process_Rank, size_Of_Cluster, ierror, message_item
integer scattered_Data
integer, dimension(4) :: tab
tab = (/21, 12, 26, 62/)

call MPI_INIT(ierror) ! Cette fonction initialise l'environnement MPI. Elle prend en compte une variable de gestion des erreurs.
call MPI_COMM_SIZE(MPI_COMM_WORLD, size_Of_Cluster, ierror) ! Cette fonction renvoie la taille totale de l'environnement en fonction de la quantité de processus. La fonction prend l'environnement MPI, un entier pour contenir le commsize, et une variable de gestion des erreurs.
call MPI_COMM_RANK(MPI_COMM_WORLD, process_Rank, ierror) !Cette fonction renvoie l'identifiant du processus qui a appelé la fonction. La fonction prend en compte l'environnement MPI, un entier pour contenir le rang de la commande, et une variable de gestion des erreurs.

 print *, 'Salut toi du process', process_Rank, 'sur ', size_Of_Cluster
! D'une execution à une autre l'autre l'affichage n'est pas séquentiel ! 


! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! Barrières et synchronisation MPI ! 
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! !  
! Comme beaucoup d'autres utilitaires de programmation parallèle, la synchronisation est un outil essentiel pour la sécurité des threads et pour s'assurer que certaines sections du code sont traitées à certains moments. MPI_BARRIER est un verrou de processus qui retient chaque processus à une certaine ligne de code jusqu'à ce que tous les processus aient atteint cette ligne.

! On affiche ici  quelque soit le rang du processus
DO i = 0, 3
        print *, 'Hey toi du process: ', process_Rank, 'sur ', size_Of_Cluster
END DO



! On affiche ici que lorsque l'itération de la boucle correspond au rang du processus.
DO i = 0, 3
    IF(i == process_Rank) THEN
        print *, 'Hey encore toi du process: ', process_Rank, 'sur ', size_Of_Cluster
    ENDIF    
END DO


! Phase de syncronisation MPI_BARRIER: cela permet de s'assurer que tous les processus sont synchronisés lors du passage dans la boucle.
DO i = 0, 3, 1
    IF(i == process_Rank) THEN
        print *, 'Oh pas mal du process: ', process_Rank, 'sur ', size_Of_Cluster
    END IF
    call MPI_BARRIER( MPI_COMM_WORLD, i_error)
END DO


! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! MPI = MESSAGE PASSING INT  ! 
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! !
! Le passage de messages est le principal utilitaire de l'interface d'application MPI qui permet aux processus de communiquer entre eux. Le passage des messages dans MPI est géré par les fonctions correspondantes et leurs arguments :  
! call MPI_SEND(integer message, integer count, MPI_Datatype datatype, integer dest,integer tag, MPI_Comm comm, integer ierror);
! call MPI_RECV(integer data, integer count, MPI_Datatype datatype, integer from,integer tag, MPI_Comm comm, MPI_Status* status, integer ierror);
! integer message              Message que vous voulez envoyer ou recevoir 
! integer count                    Nombre d'éléments envoyés dans le tableau.
! MPI_Datatype datatype   Le type de données spécifique à MPI qui est passé dans le tableau.
! integer dest                      Rang du processus de destination.
! integer from                      Rang du processus de d'envoi. 
! integer tag                        Message tag.
! MPI_Comm comm           The MPI Communicator handle.
! MPI_Status* status          Objet du statut 

IF(process_Rank == 0) THEN
    message_Item = 21
    print *, "Envoi d'un message contenant: ", message_Item
ELSE IF(process_Rank == 1) THEN
    print *, "Reception d'un message contenant: ", message_Item
END IF

! Enfin, nous devons appeler MPI_SEND() et MPI_RECV(). Nous allons passer les paramètres suivants dans les fonctions respectives 
IF(process_Rank == 0) THEN
    message_Item = 12
    call MPI_SEND(message_Item, 1, MPI_INT, 1, 1, MPI_COMM_WORLD, ierror)
    print *, "MPI_SEND Envoi d'un message contenant: ", message_Item
ELSE IF(process_Rank == 1) THEN
    call MPI_RECV(message_Item, 1, MPI_INT, 0, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierror)
    print *, "MPI_RECV Reception d'un message contenant: ", message_Item
END IF


! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! OPERATEUR DE GROUPE         ! 
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! !
! Les opérateurs de groupe sont très utiles pour MPI. Ils permettent de distribuer des bandes de données d'un processus racine à tous les autres processus disponibles, ou de collecter des données de tous les processus dans un seul processus. Ces opérateurs peuvent éliminer le besoin d'une quantité surprenante de code passe-partout via deux fonctions: MPI_Scatter et MPI_Gather
! Afin de mieux comprendre ces fonctions, nous allons créer un programme qui utilisera la fonction scatter. La fonction gather (non montrée dans l'exemple) fonctionne de manière similaire, et est essentiellement l'inverse de la fonction scatter. 
! Dans l'exemple suivant, nous allons répartir les quatre éléments d'un tableau de vecteurs entre quatre processus différents. 

call MPI_Scatter(tab, 1, MPI_INT, scattered_Data, 1, MPI_INT, 0, MPI_COMM_WORLD, ierror);
! MPI_Scatter(
!         tab,   !tableau a distribuer
!         1,              !Nombre d'élément que nous envoyons à chaque processeur
!         MPI_INT,        !MPI Datatype du tableau (evoie).
!         scattered_Data, !Variable vers laquelle est reçue la donnée 
!         1,              !Quantité de donnée transférer
!         MPI_INT,        !MPI Datatype du tableau (recoit).
!         0,              !ID du processus qui distribuera les données.
!         MPI_COMM_WORLD  !MPI Communicator.
!         ierror          !
! )
print *, "Process ", process_Rank, "recoit: ", scattered_Data
! L'exécution de ce code affichera les quatre numéros du tableau de tab comme quatre numéros distincts provenant chacun de processus différents (notez que l'ordre des rangs n'est pas nécessairement séquentiel !)

call MPI_FINALIZE(ierror)
END PROGRAM !Cette fonction nettoie l'environnement MPI et met fin aux communications MPI.