! La façon la plus simple de commencer la programmation parallèle est probablement l'utilisation d'OpenMP. OpenMP est une solution côté compilateur pour créer du code qui fonctionne sur plusieurs cœurs/threads. OpenMP étant intégré dans un compilateur, aucune bibliothèque externe ne doit être installée pour compiler ce code. 

PROGRAM OpenMP
USE OMP_LIB
INTEGER :: thread_id,tmp,Somme


! OMP_GET_THREAD_NUM() permet d'obtenir l'identifiant du processus. Cela nous permettra d'identifier chacun de nos threads en utilisant ce numéro d'identification unique.
PRINT *, 'Salut du process: ', OMP_GET_THREAD_NUM()

!  !$OMP PARALLEL et !$OMP END PARALLEL crée une section de code qui est exécutée par tous les threads disponibles, ie code à parallèle
! Avant de réaliser la parallélisation, nous devons définir le nombre de threads par la variable d'environnement : OMP_NUM_THREADS. La modification de cette variable ne nécessite pas de recompilation du programme, cette commande peut donc être placée dans la ligne de commande ou dans votre script de travail. Voir script

!$OMP PARALLEL
    PRINT *, 'Salut Parallèle du  process: ', OMP_GET_THREAD_NUM()
!$OMP END PARALLEL



! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! Variables privées et variables partagées.         !
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! La gestion de la mémoire est un élément essentiel de tout programme parallèle qui implique la manipulation de données. Dans cette extrait, nous allons voir les différents types de variables dans OpenMP ainsi qu'une implémentation simple de ces types dans le programme que nous avons fait juste au dessus
! OpenMP dispose d'une variété d'outils qui peuvent être utilisés pour indiquer correctement comment le programme parallèle doit gérer les variables. Ces outils se présentent sous la forme de classificateurs de variables partagées et privées.
! Les classificateurs privés créent une copie d'une variable pour chaque processus dans le système parallèle.
! Les classificateurs partagés détiennent une instance d'une variable pour tous les processus à partager. Pour indiquer des variables privées ou partagées, déclarez la variable avant votre section parallèle et annotez la directive omp de la manière suivante :
! !$OMP PARALLEL SHARED(shar_Var1) PRIVATE(priv_Var1, priv_Var2)
! Les variables créées et assignées à l'intérieur d'une section parallèle de code seront intrinsèquement privées, et les variables créées en dehors des sections parallèles seront intrinsèquement publiques.

!$OMP PARALLEL PRIVATE(thread_id)
    thread_id = OMP_GET_THREAD_NUM()
    PRINT *, "Coucou du process: ", thread_id
!$OMP END PARALLEL



! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! Syncronisation                                                  !
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! La directive Bareeer empêche tous les processus de passer à la ligne de code suivante jusqu'à ce que tous les processus aient atteint la barrière. Cela permet au programmeur de synchroniser les processus dans le programme parallèle.
! La directive critical garantit qu'une ligne de code n'est exécutée que par un seul processus à la fois, ce qui assure la sécurité des threads dans le corps du code.
!$OMP PARALLEL PRIVATE(thread_id)
    thread_id = OMP_GET_THREAD_NUM()

    DO i=0,OMP_GET_MAX_THREADS()
        IF (i == thread_id) THEN
            PRINT *, "Recoucou du process: ", thread_id
        END IF
        !$OMP BARRIER
    END DO
!$OMP END PARALLEL

! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! Boucle                                                               !
! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! ! 
! La puissance d'OpenMP vient du fait qu'il est facile de diviser une tâche plus importante en plusieurs tâches plus petites. Les directives de partage du travail permettent de diviser de manière simple et efficace des tâches normalement séquentielles en sections de code parallèles rapides. Dans ce qui suit, nous allons voir comment implémenter la directive omp do. La directive omp do divise une boucle for normalement séquentielle en une tâche parallèle. 

!$OMP PARALLEL PRIVATE(tmp) SHARED(Somme)
    tmp = 0;
    Somme = 0;

    !$OMP DO
    DO i=1,1000
        tmp = tmp + i
    END DO
    !$OMP END DO
    print*, 'somme calcule dans chaque thread = ', tmp
    !$OMP CRITICAL
        Somme = Somme + tmp
    !$OMP END CRITICAL

!$OMP END PARALLEL
PRINT *, "Somme: ", Somme


END